// Generated by Together
package edu.cmu.paid.learning.dataminer.developer.ID3;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;
import java.rmi.*;
import edu.cmu.paid.learning.dataminer.util.MLCDriver;
import edu.cmu.paid.learning.dataminer.developer.RemoteModelBuilder;
import edu.cmu.paid.learning.dataminer.developer.BehaviorModel;


/**
 * ID3Builder is an implementation of RemoteModelBuilder that runs as a remote
 * server that can be contacted by RemoteDeveloper for model development.
 *
 * @author Rudy Setiawan, PAID Learning Team, Carnegie Mellon University.
 * @version %I%, %G%
 */
public class ID3Builder extends java.rmi.server.UnicastRemoteObject implements RemoteModelBuilder 
{
  private String hostName;
  private String path;
  private String name = "ID3";

  /**
   * main function to start the ID3Builder as a server. The RMI registry must
   * be running before ID3Builder is started. Two arguments are needed to start
   * the ID3Builder: the RMI registry to which the ID3Builder will bind to and
   * the path where the inducer is located.
   */
  public static void main(String[] args)
  {
    String path, hostName;

    if(args.length !=2)
      return;
    else
      {
	hostName = args[0];
        path = args[1];
      }
    
    try
      {
	ID3Builder builder = new ID3Builder(hostName, path);
	builder.connect();
      }
    catch(Exception e)
      {
	System.out.println("Exception " + e + " caught.");
	e.printStackTrace();
      }
  }

  /**
   * Constructor for ID3Builder takes a couple of arguments as given below.
   *
   * @param hostName the RMI registry to which the ID3Builder will bind to
   * @param path the directory path where the ID3 inducer is located
   */
  public ID3Builder(String hostName, String path) throws java.rmi.RemoteException
    {
      this.hostName = hostName;
      this.path = path;

      if(!path.endsWith(File.separator))
	path = path + File.separator;
    }

  /** 
   * Connects to the RMI registry.
   */
  private void connect()
    { 
      try
        {
          Naming.rebind("rmi://" + hostName + "/" + name, this);

          System.out.println("ID3Builder " + name + " is ready");
        }
      catch(Exception e)
        {
          System.out.println("Exception " + e + " caught when connecting DRAgent");
        }
    }

  /**
   * Builds a BehaviorModel based on the request for model development.
   *
   * @param request the request for model development.
   * @return the BehaviorModel produced.
   * @see ID3Model
   */
  public BehaviorModel build(Hashtable request) throws java.rmi.RemoteException
    {
      System.out.println("DEBUG: ID3 Receiving request to build a model");
      Hashtable metadata = (Hashtable) request.get("metadata");
      Vector trainingSet = (Vector) request.get("training set");

      String timestamp = (new Long(System.currentTimeMillis())).toString();
      String datasetName = path + name;

      MLCDriver mlc = new MLCDriver(path, name);
      mlc.setMLCFile(metadata, trainingSet);
      String categorizerFilename = datasetName + ".id3model";
      System.out.println("DEBUG: categorizerFilename is " + categorizerFilename);
      String command = path + "ID3Inducer " + datasetName + " " + categorizerFilename;
      System.out.println("DEBUG: Command to be executed is " + command);
      mlc.runInducer(command);
      Hashtable tree = getCategorizer(metadata, categorizerFilename);

      cleanUp(datasetName);

      BehaviorModel model = new ID3Model(metadata, tree);

      System.out.println("DEBUG: Finished generating ID3Model. About to deliver");
      return model;      
    }

  /**
   * Cleans up the mess after generating the BehaviorModel.
   * 
   * @param dName the dataset's path and name
   */
  private void cleanUp(String dName)
    {
      String metaFilename = dName + ".names";
      String trainingFilename = dName + ".data";
      String categorizerFilename = dName + ".id3model";

      (new File(metaFilename)).delete();
      (new File(trainingFilename)).delete();
      (new File(categorizerFilename)).delete();
    }

  /**
   * Gets the categorizer produced by the inducer to be used in initializing
   * an instance of ID3Model.
   *
   * @param metaData the metadata of the learning task
   * @param categorizerFile the name of categorizer file to be generated by 
   *                        the inducer.
   * @return the decision tree
   */
  private Hashtable getCategorizer(Hashtable metaData, String categorizerFile)
    {
      System.out.println("Opening ID3Inducer output file");
    
      Vector rawOutput = new Vector();

      try
        {
	  FileInputStream fis = new FileInputStream(categorizerFile);
	  BufferedReader br = new BufferedReader(new InputStreamReader(fis));
	
	  while(br.ready())
	    {
	      String aLine = br.readLine();
	      rawOutput.addElement(aLine);
	    }
        }
      catch(FileNotFoundException fnfe)
        {
	  System.out.println("FileNotFoundException: " + fnfe);
        }
      catch(IOException ioe)
        {
	  System.out.println("In ID3ModelBuilder: IOException: " + ioe);
        }

      // System.out.println("rawOutput is " + rawOutput);

      Hashtable tree = new Hashtable();
    
      for(int i=3; i<rawOutput.size()-1; i++)
        {
	  String line = (String) rawOutput.elementAt(i);
	
	  // System.out.println("Line[" + i + "]: " + line);
	  StringTokenizer firstTok = new StringTokenizer(line, ":");

	  String nodeInfo = firstTok.nextToken();
	
	  String branchInfo = null;
	    
	  int branchType = -1;

	  if(firstTok.hasMoreTokens())
	    {
	      branchInfo = firstTok.nextToken();
		
	      branchInfo = branchInfo.trim();

	      if(branchInfo.length() == 0)
	        branchType = 2; // indicating leaf
	    }

	  StringTokenizer getNodeIDTok = new StringTokenizer(nodeInfo, "[]()");
	  int nodeID = Integer.valueOf(getNodeIDTok.nextToken()).intValue();
	    
	  String nodeInfo2 = getNodeIDTok.nextToken();

	  StringTokenizer getValueTok = new StringTokenizer(nodeInfo2, ",");

	  String nodeValue = getValueTok.nextToken();
	  nodeValue = nodeValue.trim();

	  String nodeInfo3 = getValueTok.nextToken();

	  StringTokenizer getLevelTok = new StringTokenizer(nodeInfo3, " ");
	
	  String junk = getLevelTok.nextToken();
	    
	  int treeLevel = Integer.valueOf(getLevelTok.nextToken()).intValue();

	  // System.out.println("NodeID: " + nodeID + "\tNode Value: " + nodeValue
	  //		   + "\tTree Level: " + treeLevel);
	
	  TreeNode nodeItem;

	  if(branchType != 2)
	    {
	      Vector possibleValues = (Vector) metaData.get(nodeValue);

	      if((possibleValues.size() == 1) && (((String) possibleValues.elementAt(0)).equals("continuous")))
	        branchType = 0;
	      else
	        branchType = 1;

	      nodeItem = new TreeNode(nodeID, branchType, nodeValue, treeLevel);

	      branchInfo = branchInfo.trim();

	      StringTokenizer getData1Tok = new StringTokenizer(branchInfo, "[]");

	      if(branchType == 0)
	        {
		  Vector boundSt = new Vector();
		  for(int j=0; getData1Tok.hasMoreTokens(); j++)
		    {
		      boundSt.addElement(getData1Tok.nextToken());
		      // System.out.println("boundSt[" + j + "] is " + boundSt.elementAt(j));
		    }
		    
		  int left, right, unknown;
		  float bound;
		
		  StringTokenizer getBoundTok;

		  if(boundSt.size() == 9)
		    {
		      unknown = Integer.valueOf((String) boundSt.elementAt(2)).intValue();
		      left = Integer.valueOf((String) boundSt.elementAt(5)).intValue();
		      right = Integer.valueOf((String) boundSt.elementAt(8)).intValue();
		      getBoundTok = new StringTokenizer((String) boundSt.elementAt(4), "-(<=) >");
		    
		      bound = Float.valueOf(getBoundTok.nextToken()).floatValue();
		      // System.out.println("Bound: " + bound + "\tUnknown: " + unknown + "\tLeft: " + left + "\tRight: " + right);
		      nodeItem.setContinuousNode(bound, left, right, unknown);
		    }
		  else
		    {
		      left = Integer.valueOf((String) boundSt.elementAt(2)).intValue();
		      right = Integer.valueOf((String) boundSt.elementAt(5)).intValue();
		      getBoundTok = new StringTokenizer((String) boundSt.elementAt(1), "-(<=) >");

		      bound = Float.valueOf(getBoundTok.nextToken()).floatValue();
		      // System.out.println("Bound: " + bound + "\tLeft: " + left + "\tRight: " + right);
		      nodeItem.setContinuousNode(bound, left, right);
		    }
	        }
	      else if (branchType == 1)
	        {
		  while(getData1Tok.hasMoreTokens())
		    {
		      junk = getData1Tok.nextToken();

		      String valueString = getData1Tok.nextToken();
		  
		      StringTokenizer getPosValueTok = new StringTokenizer(valueString, "()>");

		      junk = getPosValueTok.nextToken();

		      int nextID = Integer.valueOf(getData1Tok.nextToken()).intValue();
		      String nextValue = getPosValueTok.nextToken();
			
		      nodeItem.addDiscreteBranch(nextValue, nextID);

		      // System.out.println("For value " + nextValue + ", go to " + nextID);
		    }
	        }
	    }
	  else
	    {
	      nodeItem = new TreeNode(nodeID, branchType, nodeValue, treeLevel);
	    }

	  tree.put(new Integer(nodeID), nodeItem);
        }
	
      // System.out.println("DEBUG: Tree is " + tree);

      return tree;
    }
}
